"""
Pythonic interface to SLEIGH
"""

from __future__ import annotations
import os.path
import xml.etree.ElementTree as ET
from enum import IntEnum
from typing import cast, Generator, Sequence, Mapping

from .__version__ import __version__

from .pypcode_native import __version__ as __pypcode_native_version__  # pylint:disable=no-name-in-module

assert __version__ == __pypcode_native_version__

from .pypcode_native import (  # pylint:disable=no-name-in-module
    Address,
    AddrSpace,
    BadDataError,
    DecoderError,
    Disassembly,
    Instruction,
    LowlevelError,
    OpCode,
    PcodeOp,
    TRANSLATE_FLAGS_BB_TERMINATING,
    Translation,
    UnimplError,
    Varnode,
)
from .pypcode_native import Context as _Context  # pylint:disable=no-name-in-module

from .printing import (
    OpFormat,
    OpFormatBinary,
    OpFormatFunc,
    OpFormatSpecial,
    OpFormatUnary,
    PcodePrettyPrinter,
)

__all__ = [
    "Address",
    "AddrSpace",
    "Arch",
    "ArchLanguage",
    "BadDataError",
    "Context",
    "DecoderError",
    "Disassembly",
    "Instruction",
    "LowlevelError",
    "OpCode",
    "OpFormat",
    "OpFormatBinary",
    "OpFormatFunc",
    "OpFormatSpecial",
    "OpFormatUnary",
    "PcodeOp",
    "PcodePrettyPrinter",
    "TranslateFlags",
    "Translation",
    "UnimplError",
    "Varnode",
]


class TranslateFlags(IntEnum):
    """
    Flags that can be passed to Context::translate
    """

    BB_TERMINATING = TRANSLATE_FLAGS_BB_TERMINATING


PKG_SRC_DIR = os.path.abspath(os.path.dirname(__file__))
SPECFILES_DIR = os.path.join(PKG_SRC_DIR, "processors")


class ArchLanguage:
    """
    A specific language for an architecture. Provides access to language, pspec, and cspecs.
    """

    __slots__ = (
        "archdir",
        "ldef",
        "_pspec",
        "_cspecs",
    )

    archdir: str
    ldef: ET.Element

    def __init__(self, archdir: str, ldef: ET.Element):
        self.archdir = archdir
        self.ldef = ldef
        self._pspec: ET.Element | None = None
        self._cspecs: dict[tuple[str, str], ET.Element] | None = None

    @property
    def pspec_path(self) -> str:
        return os.path.join(self.archdir, self.processorspec)

    @property
    def slafile_path(self) -> str:
        return os.path.join(self.archdir, self.slafile)

    @property
    def description(self) -> str:
        elem = self.ldef.find("description")
        if elem is not None:
            return elem.text or ""
        return ""

    def __getattr__(self, key):
        if key in self.ldef.attrib:
            return self.ldef.attrib[key]
        raise AttributeError(key)

    @property
    def pspec(self) -> ET.Element | None:
        if self._pspec is None:
            self._pspec = ET.parse(self.pspec_path).getroot()
        return self._pspec

    @property
    def cspecs(self) -> Mapping[tuple[str, str], ET.Element]:
        if self._cspecs is None:
            self._cspecs = {}
            for e in self.ldef.findall("compiler"):
                path = os.path.join(self.archdir, e.attrib["spec"])
                cspec = ET.parse(path).getroot()
                self._cspecs[(e.attrib["id"], e.attrib["name"])] = cspec
        return self._cspecs

    def init_context_from_pspec(self, ctx: Context) -> None:
        if self.pspec is None:
            return
        cd = self.pspec.find("context_data")
        if cd is None:
            return
        cs = cd.find("context_set")
        if cs is None:
            return
        for e in cs:
            assert e.tag == "set"
            ctx.setVariableDefault(e.attrib["name"], int(e.attrib["val"]))

    @classmethod
    def from_id(cls, langid: str) -> ArchLanguage | None:
        """
        Return language with given id, or None if the language could not be found.
        """
        for arch in Arch.enumerate():
            for lang in arch.languages:
                if lang.id == langid:
                    return lang
        return None


class Arch:
    """
    Main class representing an architecture describing available languages.
    """

    __slots__ = (
        "archpath",
        "archname",
        "ldefpath",
        "ldef",
        "languages",
    )

    archpath: str
    archname: str
    ldefpath: str
    ldef: ET.ElementTree[ET.Element[str]]
    languages: Sequence[ArchLanguage]

    def __init__(self, name: str, ldefpath: str):
        """
        Initialize the Arch.

        :param name: The name of the architecture
        :param ldefpath: Path to language definition files (.ldefs)
        """
        self.archpath = os.path.dirname(ldefpath)
        self.archname = name
        self.ldefpath = ldefpath
        self.ldef = ET.parse(ldefpath)
        self.languages = [ArchLanguage(self.archpath, e) for e in self.ldef.getroot()]

    @classmethod
    def enumerate(cls) -> Generator[Arch]:
        """
        Enumerate all available architectures and languages.

        Language definitions are sourced from definitions shipped with pypcode and
        can be found in processors/<architecture>/data/languages/<variant>.ldefs
        """
        for archname in os.listdir(SPECFILES_DIR):
            langdir = os.path.join(SPECFILES_DIR, archname, "data", "languages")
            if not (os.path.exists(langdir) and os.path.isdir(langdir)):
                continue
            for langname in os.listdir(langdir):
                if not langname.endswith(".ldefs"):
                    continue
                ldefpath = os.path.join(langdir, langname)
                yield Arch(archname, ldefpath)


class Context(_Context):
    """
    Context for translation.
    """

    __slots__ = (
        "language",
        "registers",
    )

    language: ArchLanguage
    registers: dict[str, Varnode]

    def __init__(self, language: ArchLanguage | str):
        """
        Initialize a context for translation or disassembly.

        :param language: The ``ArchLanguage`` to initialize the context with, or a language id ``str``
        """
        if isinstance(language, ArchLanguage):
            self.language = language
        elif isinstance(language, str):
            _l = ArchLanguage.from_id(cast(str, language))
            assert _l is not None
            self.language = _l
        else:
            raise TypeError("Context must be initialized with a language or language id")
        super().__init__(f"<sleigh>{self.language.slafile_path}</sleigh>")
        self.language.init_context_from_pspec(self)
        self.registers = {n: v for v, n in self.getAllRegisters().items()}



#!/usr/bin/env python
"""
Runs when invoking pypcode module from command line. Lists supported
architectures, and handles basic disassembly and translation to P-code of
supported binaries. Does not parse object files, the binary files must be plain
machine code bytes in a file.
"""

import argparse
import logging
import sys
from difflib import SequenceMatcher

from pypcode import Arch, BadDataError, Context, OpCode, TranslateFlags, UnimplError

log = logging.getLogger(__name__)
logging.basicConfig(level=logging.DEBUG, format="[%(name)s:%(levelname)s] %(message)s")


def main():
    ap = argparse.ArgumentParser(
        prog="pypcode",
        description="Disassemble and translate machine code to P-code using SLEIGH",
    )
    ap.add_argument(
        "-l",
        "--list",
        action="store_true",
        help="list supported architecture languages",
    )
    ap.add_argument("langid", help="architecture language id")
    ap.add_argument("binary", help="path to flat binary code")
    ap.add_argument("base", default="0", nargs="?", help="base address to load at")
    ap.add_argument("-o", "--offset", default="0", help="offset in binary file to load from")
    ap.add_argument("-s", "--length", default=None, help="length of code in bytes to load")
    ap.add_argument(
        "-i",
        "--max-instructions",
        default=0,
        type=int,
        help="maximum number of instructions to translate",
    )
    ap.add_argument(
        "-b",
        "--basic-block",
        action="store_true",
        default=False,
        help="stop translation at end of basic block",
    )

    # List supported languages
    langs = {lang.id: lang for arch in Arch.enumerate() for lang in arch.languages}
    if ("-l" in sys.argv) or ("--list" in sys.argv):
        for langid in sorted(langs):
            print("%-35s - %s" % (langid, langs[langid].description))
        return

    args = ap.parse_args()

    # Get requested language
    if args.langid not in langs:
        print(f'Language "{args.langid}" not found.')
        t = args.langid.upper()
        suggestions = [langid for langid in langs if SequenceMatcher(None, t, langid.split()[0].upper()).ratio() > 0.25]
        if len(suggestions):
            print("\nSuggestions:")
            for langid in sorted(suggestions):
                print("  %-35s - %s" % (langid, langs[langid].description))
            print("")
        print("Try `--list` for full list of architectures.")
        sys.exit(1)

    # Load target binary code
    base = int(args.base, 0)
    with open(args.binary, "rb") as f:
        f.seek(int(args.offset, 0))
        code = f.read(int(args.length, 0)) if args.length else f.read()

    # Translate
    ctx = Context(langs[args.langid])

    try:
        flags = TranslateFlags.BB_TERMINATING if args.basic_block else 0
        res = ctx.translate(code, base, max_instructions=args.max_instructions, flags=flags)

        last_imark_idx = 0
        for i, op in enumerate(res.ops):
            if op.opcode == OpCode.IMARK:
                last_imark_idx = i
                disas_addr = op.inputs[0].offset
                disas_offset = disas_addr - base
                disas_len = sum(vn.size for vn in op.inputs)
                disas_slice = code[disas_offset : disas_offset + disas_len]
                print(ctx.disassemble(disas_slice, disas_addr))
            else:
                print(f" {i - last_imark_idx - 1:3d}: {op}")
        print("")
    except (BadDataError, UnimplError) as e:
        print(f"An error occurred during translation: {e}")
        sys.exit(1)


if __name__ == "__main__":
    main()



from typing import Any, ClassVar, List, Optional

__version__: str

TRANSLATE_FLAGS_BB_TERMINATING: int

class BadDataError(Exception): ...
class DecoderError(Exception): ...
class LowlevelError(Exception): ...
class UnimplError(Exception): ...

class AddrSpace:
    def __init__(self, *args, **kwargs) -> None: ...
    @property
    def name(self) -> str: ...

class Address:
    def __init__(self, *args, **kwargs) -> None: ...
    @property
    def offset(self) -> int: ...
    @property
    def space(self) -> AddrSpace: ...

class Instruction:
    def __init__(self, *args, **kwargs) -> None: ...
    @property
    def addr(self) -> Address: ...
    @property
    def body(self) -> str: ...
    @property
    def length(self) -> int: ...
    @property
    def mnem(self) -> str: ...

class Disassembly:
    def __init__(self, *args, **kwargs) -> None: ...
    @property
    def instructions(self) -> List[Instruction]: ...

class OpCode:
    BOOL_AND: ClassVar[OpCode] = ...
    BOOL_NEGATE: ClassVar[OpCode] = ...
    BOOL_OR: ClassVar[OpCode] = ...
    BOOL_XOR: ClassVar[OpCode] = ...
    BRANCH: ClassVar[OpCode] = ...
    BRANCHIND: ClassVar[OpCode] = ...
    CALL: ClassVar[OpCode] = ...
    CALLIND: ClassVar[OpCode] = ...
    CALLOTHER: ClassVar[OpCode] = ...
    CAST: ClassVar[OpCode] = ...
    CBRANCH: ClassVar[OpCode] = ...
    COPY: ClassVar[OpCode] = ...
    CPOOLREF: ClassVar[OpCode] = ...
    EXTRACT: ClassVar[OpCode] = ...
    FLOAT_ABS: ClassVar[OpCode] = ...
    FLOAT_ADD: ClassVar[OpCode] = ...
    FLOAT_CEIL: ClassVar[OpCode] = ...
    FLOAT_DIV: ClassVar[OpCode] = ...
    FLOAT_EQUAL: ClassVar[OpCode] = ...
    FLOAT_FLOAT2FLOAT: ClassVar[OpCode] = ...
    FLOAT_FLOOR: ClassVar[OpCode] = ...
    FLOAT_INT2FLOAT: ClassVar[OpCode] = ...
    FLOAT_LESS: ClassVar[OpCode] = ...
    FLOAT_LESSEQUAL: ClassVar[OpCode] = ...
    FLOAT_MULT: ClassVar[OpCode] = ...
    FLOAT_NAN: ClassVar[OpCode] = ...
    FLOAT_NEG: ClassVar[OpCode] = ...
    FLOAT_NOTEQUAL: ClassVar[OpCode] = ...
    FLOAT_ROUND: ClassVar[OpCode] = ...
    FLOAT_SQRT: ClassVar[OpCode] = ...
    FLOAT_SUB: ClassVar[OpCode] = ...
    FLOAT_TRUNC: ClassVar[OpCode] = ...
    IMARK: ClassVar[OpCode] = ...
    INDIRECT: ClassVar[OpCode] = ...
    INSERT: ClassVar[OpCode] = ...
    INT_2COMP: ClassVar[OpCode] = ...
    INT_ADD: ClassVar[OpCode] = ...
    INT_AND: ClassVar[OpCode] = ...
    INT_CARRY: ClassVar[OpCode] = ...
    INT_DIV: ClassVar[OpCode] = ...
    INT_EQUAL: ClassVar[OpCode] = ...
    INT_LEFT: ClassVar[OpCode] = ...
    INT_LESS: ClassVar[OpCode] = ...
    INT_LESSEQUAL: ClassVar[OpCode] = ...
    INT_MULT: ClassVar[OpCode] = ...
    INT_NEGATE: ClassVar[OpCode] = ...
    INT_NOTEQUAL: ClassVar[OpCode] = ...
    INT_OR: ClassVar[OpCode] = ...
    INT_REM: ClassVar[OpCode] = ...
    INT_RIGHT: ClassVar[OpCode] = ...
    INT_SBORROW: ClassVar[OpCode] = ...
    INT_SCARRY: ClassVar[OpCode] = ...
    INT_SDIV: ClassVar[OpCode] = ...
    INT_SEXT: ClassVar[OpCode] = ...
    INT_SLESS: ClassVar[OpCode] = ...
    INT_SLESSEQUAL: ClassVar[OpCode] = ...
    INT_SREM: ClassVar[OpCode] = ...
    INT_SRIGHT: ClassVar[OpCode] = ...
    INT_SUB: ClassVar[OpCode] = ...
    INT_XOR: ClassVar[OpCode] = ...
    INT_ZEXT: ClassVar[OpCode] = ...
    LOAD: ClassVar[OpCode] = ...
    MULTIEQUAL: ClassVar[OpCode] = ...
    NEW: ClassVar[OpCode] = ...
    PIECE: ClassVar[OpCode] = ...
    POPCOUNT: ClassVar[OpCode] = ...
    LZCOUNT: ClassVar[OpCode] = ...
    PTRADD: ClassVar[OpCode] = ...
    PTRSUB: ClassVar[OpCode] = ...
    RETURN: ClassVar[OpCode] = ...
    SEGMENTOP: ClassVar[OpCode] = ...
    STORE: ClassVar[OpCode] = ...
    SUBPIECE: ClassVar[OpCode] = ...
    __entries: ClassVar[dict] = ...
    __name__: Any
    def __init__(self, *args, **kwargs) -> None: ...
    def __eq__(self, other) -> Any: ...
    def __ge__(self, other) -> Any: ...
    def __gt__(self, other) -> Any: ...
    def __hash__(self) -> Any: ...
    def __index__(self) -> Any: ...
    def __int__(self) -> Any: ...
    def __le__(self, other) -> Any: ...
    def __lt__(self, other) -> Any: ...
    def __ne__(self, other) -> Any: ...

class Varnode:
    def __init__(self) -> None: ...
    def getRegisterName(self) -> str: ...
    def getUserDefinedOpName(self) -> str: ...
    def getSpaceFromConst(self) -> AddrSpace: ...
    @property
    def offset(self) -> int: ...
    @property
    def size(self) -> int: ...
    @property
    def space(self) -> AddrSpace: ...

class PcodeOp:
    def __init__(self, *args, **kwargs) -> None: ...
    @property
    def inputs(self) -> List[Varnode]: ...
    @property
    def opcode(self) -> OpCode: ...
    @property
    def output(self) -> Optional[Varnode]: ...

class Translation:
    def __init__(self, *args, **kwargs) -> None: ...
    @property
    def ops(self) -> List[PcodeOp]: ...

class Context:
    def __init__(self, *args, **kwargs) -> None: ...
    def disassemble(
        self,
        bytes: bytes,
        base_address: int = ...,
        offset: int = ...,
        max_bytes: int = ...,
        max_instructions: int = ...,
    ) -> Disassembly: ...
    def getAllRegisters(self) -> dict[Varnode, str]: ...
    def getRegisterName(self, space: AddrSpace, offset: int, size: int) -> str: ...
    def reset(self) -> None: ...
    def setVariableDefault(self, name: str, value: int) -> None: ...
    def translate(
        self,
        bytes: bytes,
        base_address: int = ...,
        offset: int = ...,
        max_bytes: int = ...,
        max_instructions: int = ...,
        flags: int = ...,
    ) -> Translation: ...
